"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.getOutDirAbsolute = exports.determineOutDir = void 0;
const assert_1 = require("./assert");
const filesystemPathHandling_1 = require("./filesystemPathHandling");
const path_1 = __importDefault(require("path"));
/** Appends `client/` or `server/` to `config.build.outDir` */
function determineOutDir(config) {
    // The mechansism to detect a framework down below doesn't work for Astro
    //  - https://github.com/withastro/astro/issues/5211#issuecomment-1326084151
    if (config.plugins.some((p) => { var _a; return (_a = p.name) === null || _a === void 0 ? void 0 : _a.startsWith('astro:'); }))
        return null;
    const outDirRoot = (0, filesystemPathHandling_1.toPosixPath)(config.build.outDir);
    (0, filesystemPathHandling_1.assertPosixPath)(outDirRoot);
    // Mechanism to detect whether Telefunc is used with framework.
    // When used with a framework then Telefunc should let the framework determine `outDir`.
    // E.g. Vike and SvelteKit already set `config.build.outDir`.
    if (!isOutDirRoot(outDirRoot)) {
        assertConfig(config);
        return null;
    }
    const { outDirClient, outDirServer } = declineOutDirs(outDirRoot);
    if (config.build.ssr) {
        return outDirServer;
    }
    else {
        return outDirClient;
    }
}
exports.determineOutDir = determineOutDir;
function declineOutDirs(outDirRoot) {
    assertIsOutDirRoot(outDirRoot);
    (0, filesystemPathHandling_1.assertPosixPath)(outDirRoot);
    const outDirClient = path_1.default.posix.join(outDirRoot, 'client');
    const outDirServer = path_1.default.posix.join(outDirRoot, 'server');
    assertIsNotOutDirRoot(outDirClient);
    assertIsNotOutDirRoot(outDirServer);
    return { outDirClient, outDirServer };
}
function assertIsOutDirRoot(outDir) {
    (0, assert_1.assert)(isOutDirRoot(outDir));
}
function isOutDirRoot(outDir) {
    const p = outDir.split('/').filter(Boolean);
    const lastDir = p[p.length - 1];
    return lastDir !== 'client' && lastDir !== 'server';
}
function assertIsNotOutDirRoot(outDir) {
    (0, assert_1.assert)(outDir.endsWith('/client') || outDir.endsWith('/server'));
}
function assertConfig(config) {
    var _a;
    const outDir = (_a = config.build) === null || _a === void 0 ? void 0 : _a.outDir;
    (0, assert_1.assert)(outDir);
    assertIsNotOutDirRoot(outDir);
    if (config.build.ssr) {
        (0, assert_1.assert)(outDir.endsWith('/server'));
    }
    else {
        (0, assert_1.assert)(outDir.endsWith('/client'));
    }
}
function getOutDirAbsolute(config) {
    let { outDir } = config.build;
    // SvelteKit sets config.build.outDir to `D:\a\telefunc\telefunc\examples\svelte-kit\.svelte-kit/output/server.`
    outDir = (0, filesystemPathHandling_1.toPosixPath)(outDir);
    if (!outDirIsAbsolutePath(outDir)) {
        const { root } = config;
        (0, filesystemPathHandling_1.assertPosixPath)(root);
        outDir = path_1.default.posix.join(root, outDir);
    }
    return outDir;
}
exports.getOutDirAbsolute = getOutDirAbsolute;
function outDirIsAbsolutePath(outDir) {
    // There doesn't seem to be a better alternative to determine whether `outDir` is an aboslute path
    //  - Very unlikely that `outDir`'s first dir macthes the filesystem's first dir
    return getFirstDir(outDir) === getFirstDir(process.cwd());
}
function getFirstDir(p) {
    const firstDir = p.split(/\/|\\/).filter(Boolean)[0];
    return firstDir;
}
